require 'cocoapods'
require 'set'

module Pod

  class Image
    attr_accessor :dir
    attr_accessor :name
    attr_accessor :fuzzy_name
    def initialize(dir, name)
      @dir = dir
      if name.end_with?('@2x') || name.end_with?('@3x')
        name = name[0, name.length-3]
      end
      @name = name

      # 只处理最后一段为数字，或分了三段及以上的命名形式
      comps = name.split('_')
      last = comps.last
      if last.to_i > 0 || last.start_with?('0') || comps.count >= 3
        @fuzzy_name = name[0, name.length - last.length]
      else
        @fuzzy_name = ''
      end
    end

    def eql?(other)
      @dir == other.dir && @name == other.name
    end

    def hash
      [@dir, @name].hash
    end

    def fullpath
      File.join(@dir, @name)
    end
  end

  class Command

    class Strip < Command

      self.summary = '查找工程内所有无效资源'

      self.description = <<-DESC
        查找工程内所有无效资源
      DESC

      def validate!
        super

        @img_exts = ['.png', '.jpg', '.jpeg', '.webp', '.gif']

        @proj_paths = []
        @proj_paths |= Pathname.glob('*.xcodeproj')
        @proj_paths |= Pathname.glob('Pods/*.xcodeproj')

        help! '未发现任何工程' unless @proj_paths.count > 0
      end

      # 遍历目录找出所有图片
      def traverse(path, except_dirs, bk = nil)
        dir = File.dirname(path)
        ext = File.extname(path)
        name = File.basename(path, ext)
        basename = File.basename(path)
        if except_dirs.include?(basename)
          return
        end
        if File.directory?(path)
          if ext == '.imageset'
            # imageset 直接返回
            bk.call(dir, name, ext)
          else
            Dir.foreach(path) do |name|
              unless name.start_with?('.') || name.end_with?('.launchimage')
                traverse(File.join(path, name), except_dirs, bk)
              end
            end
          end
        else
          bk.call(dir, name, ext) if @img_exts.include?(ext)
        end
      end

      # 获取源文件
      def get_source_files
        source_files = []
        @proj_paths.each do |path|
          proj = Xcodeproj::Project.open(path)
          proj.targets.each do |target| 
            target.build_phases.each do |phase|
              if phase.is_a?(Xcodeproj::Project::PBXSourcesBuildPhase)
                source_files |= phase.files
              end
            end
          end
        end
        source_files
      end

      # 获取所有图片资源
      def get_images
        images = Set[]
        except_dirs = [
          'QYResource.bundle',
          'DJMate',
          'IQKeyboardManager',
          'MJRefresh.bundle',
          'AlipaySDK.bundle',
          'AppIcon.appiconset',
          'KPCameraImages.xcassets',
          'UMSocialSDKResources.bundle',
          'LaunchResource'
        ]
        except_files = [
          'image_placeholder',
          'register_add_avatar'
        ]
        bk = Proc.new do |dir, name, ext|
          next if except_files.include?(name)
          images << Image.new(dir, name)
        end
        traverse(Dir.pwd, except_dirs, bk)
        images
      end

      # 分析
      def analyze(source_files, images)
        source_file_paths = source_files.map { |x| 
          x.file_ref.real_path
        }
        .select { |x| 
          File.exist?(x)
        }
        count = source_file_paths.count
        source_file_paths.each_index do | index |
          percent = index.to_f / count.to_f * 100
          print "\r" if index != 0
          print "#{format("%.2f", percent)}%"
          path = source_file_paths[index]
          File.open(path, 'r') do |f|
            image_using = []
            f.each_line do |line|
              next unless line.include?('"')
              ['"', '/'].each do |prefix|
                images.each do |img|
                  if line.include?(prefix + img.name)
                    image_using << img
                  else
                    # 根据下划线分割剔除最后一项，再尝试匹配
                    if !img.fuzzy_name.empty?
                      ['%', '"'].each do |suffix|
                        image_using << img if line.include?(prefix + img.fuzzy_name + suffix)
                      end
                    end
                  end
                end
              end
            end
            images = images - image_using
          end
        end
        print "\r"
        images
      end

      def run
        images = analyze(get_source_files, get_images)
        if images.empty?
          puts "未找到无效资源"
        else
          puts "无效资源文件："
          images.each do |img|
            path = Pathname.new(img.fullpath).relative_path_from(Dir.pwd).to_s
            puts "  #{path}"
          end
        end
      end
    end
  end
end