require 'xcodeproj'
require 'fileutils'
require 'dongjia_config'
require 'time'

module Dongjia

  HEADER_CONTENT = <<-EOS
/**

  这里汇总了项目中所有的路由定义，方便日常开发中进行检索、版本比对等操作

  该文件会在每次 pod install 后，由 cocoapods-dongjia 插件自动生成，所以无需手动修改

  注意：此处的实现仅供查看，如需修改，请点击对应的 _redirect{type} 方法跳转至对应的实现

  跳转定义文档：https://cf.idongjia.cn/pages/viewpage.action?pageId=57351032
  
*/


EOS

  class RouterItem
    
    attr_accessor :body
    attr_accessor :redirect_type
    attr_accessor :pre_line

    def initialize(defining_line)
      @defining_line = defining_line.strip
      @body = ''
      @prefix = nil
      prefix_len = 0
      if @defining_line.start_with?('@RedirectObj(') || @defining_line.start_with?('@redirectObj(')
        @prefix = 'RedirectObj'
        prefix_len = @prefix.length + 2
      elsif @defining_line.start_with?('@Redirect(') || @defining_line.start_with?('@redirect(')
        @prefix = 'Redirect'
        prefix_len = @prefix.length + 2
      end
      
      if @prefix != nil
        end_index = @defining_line.index(',') || @defining_line.index(')')
        if !end_index.nil?
          @redirect_type = @defining_line[prefix_len, end_index - prefix_len]
        end
      end
    end

    def output
      del_count = @defining_line.end_with?('{') ? 1 : 0
      result = "#pregma mark - "
      result << @defining_line[0, @defining_line.length - del_count].strip

      if @pre_line && @pre_line.start_with?('/')
        result << " " << @pre_line.gsub(/^\/+/, '').strip
      end

      result << "\n"
      result << "_" << @prefix << @redirect_type << "\n"
      result << (del_count == 1 ? "{\n" : "")
      result << @body << "\n\n"
      result
    end
  end

  class Router

    def route_items_from_file(path)
      route_items = []
      item = nil
      File.open(path, 'r') do |file|
        pre_line = nil
        file.each_line do |line|
          is_start = false
          if line.start_with?('@Redirect') || line.start_with?('@redirect')
            # 路由定义开始
            item = RouterItem.new(line)
            item.pre_line = pre_line
            is_start = true
          elsif item != nil && line.start_with?('}')
            # 路由定义结束
            item.body << line
            route_items << item
            item = nil
            pre_line = nil
          end
          if item != nil
            if !is_start
              item.body << line
            end
          end
          pre_line = line
        end
      end
      route_items
    end

    # 开始抓取
    def start_scrapy(sandbox_root, result_file_path)
      
      route_items = []

      Pod::UI.puts("Scraping routers")

      Dir.foreach(sandbox_root).select{|f| f.end_with?('xcodeproj')}.each do |name|
        proj = Xcodeproj::Project.open(File.join(sandbox_root, name))
        if name != 'Pods.xcodeproj'
          proj.targets.each do |target|
            next unless target.name.start_with?('DJ')
            next unless target.is_a?(Xcodeproj::Project::Object::PBXNativeTarget)
            target.source_build_phase.files_references.each do |file|
              route_items += route_items_from_file(file.real_path)
            end
          end
        end
      end

      # 排序
      route_items.sort! do |a, b|
        a.redirect_type.to_i <=> b.redirect_type.to_i
      end

      # 输出所有结果
      result = HEADER_CONTENT
      route_items.each do |item|
        result += item.output
      end

      FileUtils.chmod('u+w', result_file_path)

      File.open(result_file_path, 'w') do |f|
        f.write(result)
      end

      # 将文件只读
      FileUtils.chmod('ugo-w', result_file_path)
    end

    # 外部入口
    def scrape_routers(sandbox_root, config)

      if config == nil
        return
      end

      begin

        result_file_path = ''
        path = config[:path]
        if path && path.length > 0
          result_file_path = File.expand_path(File.join(sandbox_root, '..', path))
        end
        
        if !File.exist?(result_file_path)
          Pod::UI.warn("Router scrapy: path not found.")
          return
        end

        key = 'latest_scrapy_router_tm'
        content = File.read(result_file_path)
        if content.length < 10
          Config.check do |cfg, save| 
            start_scrapy(sandbox_root, result_file_path)
            cfg[key] = Time.now.to_i
            save.call
          end
        else
          Config.is_expired?(key, 24 * 60 * 60) do |_, update|
            start_scrapy(sandbox_root, result_file_path)
            update.call
          end
        end

      rescue => e
        puts "Error: #{e}"
      end

    end

  end

end